@@ -58,12 +58,12 @@
 #include <stdarg.h>
 #include <assert.h>
 
-#include <ssl.h>
-#include <err.h>
-#include <x509.h>
-#include <pem.h>
-#include <crypto.h>
-#include <rand.h>
+#include <openssl/ssl.h>
+#include <openssl/err.h>
+#include <openssl/x509.h>
+#include <openssl/pem.h>
+#include <openssl/crypto.h>
+#include <openssl/rand.h>
 
 #include "httpd.h"
 #include "http_config.h"
@@ -71,10 +71,6 @@
 #include "http_main.h"
 
 
-#if SSLEAY_VERSION_NUMBER < 0x0920
-# error "Don't use OpenSSL versions less than 0.9.2b, they have a serious security problem!"
-#endif
-
 /* Note that APACHE_SSL_EXPORT_CERTS is defined in buff.h */
 
 /* Temp */
@@ -84,7 +80,9 @@
 
 #define USE_OLD_ENVIRONMENT	FALSE	/* Set to TRUE to use the old SSL_{CLIENT,SERVER}_I* instead of SSL_{CLIENT,SERVER}_I_* */
 
-#define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
+#define SSL_IS_OPTIONAL		TRUE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
+
+#define USE_EXTENDED_ENVIRONMENT      TRUE    /* Set to TRUE to use the SSL_{CLIENT,SERVER}_{I,S}_DN* */
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
 /* Do session caching - only available with SSLeay 0.8.x */
@@ -145,6 +143,14 @@ static char *s_szCacheServerPath;
 # define X509_verify_cert_error_string(error) X509_cert_verify_error_string(error)
 #endif
 
+#if OPENSSL_VERSION_NUMBER >= 0x00904000L
+# define localPEM_read_RSAPrivateKey(f,k,x) PEM_read_RSAPrivateKey(f,k,x,NULL)
+# define localPEM_read_X509(f,c,x)	    PEM_read_X509(f,c,x,NULL)
+#else
+# define localPEM_read_RSAPrivateKey(f,k,x) PEM_read_RSAPrivateKey(f,k,x)
+# define localPEM_read_X509(f,c,x)	    PEM_read_X509(f,c,x)
+#endif
+
 #define ENV_CERT_CHAIN		"SSL_CLIENT_CERT_CHAIN_"
 
 #define UNSET	-1
@@ -165,6 +171,13 @@ typedef enum
     VERIFY_OPTIONAL_NO_CA=3
     } VerifyType;
 
+typedef enum
+    {
+    SSLREQ_UNSET=-1,
+    SSLREQ_REQUIRED=0,
+    SSLREQ_DENIED=1
+    } RequireType;
+
 typedef struct
     {
     BOOL bDisabled;
@@ -200,7 +213,7 @@ typedef struct
 
     table *tbRequiredCiphers;
     table *tbBannedCiphers;
-    BOOL bSSLRequired;
+    RequireType nSSLRequired;
 #if APACHE_SSL_EXPORT_CERTS
     BOOL bExportCerts;
 #endif
@@ -425,7 +438,11 @@ static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
     if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
        || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
        || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
-       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
+       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN
+#if OPENSSL_VERSION_NUMBER >= 0x00905000L
+       || error == X509_V_ERR_CERT_UNTRUSTED
+#endif
+       )
 	{
 	char *s;
 
@@ -495,13 +512,19 @@ static int VerifyFlags(SSLConfigRec *pConfig)
 
 static int SSLCheckCipher(request_rec *r)
     {
-    char *cipher;
+    const char *cipher;
     SSLDirConfigRec *rec=(SSLDirConfigRec *)
       ap_get_module_config(r->per_dir_config,&apache_ssl_module);
     SSLConfigRec *pConfig=(SSLConfigRec *)
       ap_get_module_config(r->server->module_config,&apache_ssl_module);
 
-    if(rec->bSSLRequired && !r->connection->client->ssl)
+    if(rec->nSSLRequired == SSLREQ_DENIED && r->connection->client->ssl)
+	{
+	ap_log_reason("SSL denied",r->filename,r);
+	return FORBIDDEN;
+	}
+
+    if(rec->nSSLRequired == SSLREQ_REQUIRED && !r->connection->client->ssl)
 	{
 	ap_log_reason("SSL required",r->filename,r);
 	return FORBIDDEN;
@@ -511,6 +534,8 @@ static int SSLCheckCipher(request_rec *r)
     if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
+    if (r->connection->client->ssl == NULL) return DECLINED;
+
 #if RENEG
 # if DEBUG_RENEG
     fprintf(stderr,"Renegotiated for %s (%x)!\n",r->filename,VerifyFlags(rec));
@@ -520,6 +545,11 @@ static int SSLCheckCipher(request_rec *r)
 	SSL_set_ex_data(r->connection->client->ssl,EX_DATA_DIR_CONFIG,rec);
 	SSL_set_verify(r->connection->client->ssl,VerifyFlags(rec),
 		       ApacheSSLVerifyCallback);
+	/* Set the context. This is probably not optimal (but nor is the whole,
+	   of renegotiation. But it certainly should avoid inappropriate
+	   session reuse. */
+	SSL_set_session_id_context(r->connection->client->ssl,
+				   (unsigned char *)&rec,sizeof rec);
 	SSL_renegotiate(r->connection->client->ssl);
 	SSL_do_handshake(r->connection->client->ssl);
 
@@ -582,7 +612,42 @@ static void StrUpper(char *s)
 	*s=toupper(*s);
     }
 
-static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
+#ifdef USE_EXTENDED_ENVIRONMENT
+static void get_serial(pool *p,table *pEnv,char *name, X509 *xs)
+{
+    BIO *bio;
+    char *result;
+    int n;
+
+    if ((bio = BIO_new(BIO_s_mem())) == NULL) return;
+    i2a_ASN1_INTEGER(bio, X509_get_serialNumber(xs));
+    n = BIO_pending(bio);
+    result = ap_pcalloc(p, n+1);
+    n = BIO_read(bio, result, n);
+    result[n] = 0;
+    ap_table_setn(pEnv,name, result);
+    BIO_free(bio);
+}
+
+static void get_valid(pool *p, table *pEnv, char *name, ASN1_UTCTIME *tm)
+{
+    BIO *bio;
+    char *result;
+    int n;
+
+    if ((bio = BIO_new(BIO_s_mem())) == NULL) return;
+    ASN1_UTCTIME_print(bio, tm);
+    n = BIO_pending(bio);
+    result = ap_pcalloc(p, n+1);
+    n = BIO_read(bio, result, n);
+    result[n] = 0;
+    ap_table_setn(pEnv, name, result);
+    BIO_free(bio);
+}
+
+#endif
+
+static void ExpandCert(pool *p,table *pEnv,char *szPrefix, char *szDN, char *szCert)
     {
     char buf[HUGE_STRING_LEN];
     char *s,*t;
@@ -592,8 +657,7 @@ static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
        with CGI's written for stronghold's certificate info  - Q */
     /* FIXME - strtok() and strcspn() may cause problems on some systems - Q */
 
-    t=ap_psprintf(p,"%sDN",szPrefix);
-    ap_table_setn(pEnv,t,ap_pstrdup(p,szCert));
+    ap_table_setn(pEnv,szDN,ap_pstrdup(p,szCert));
     
     ap_cpystrn(buf,szCert,sizeof buf);
     for(s=strtok(buf,"/") ; s != NULL ; s=strtok(NULL,"/"))
@@ -655,13 +719,13 @@ static struct
     { NULL, 0, 0 }
     };
 
-static int SSLFixups(request_rec *r)
+static void SSLAddCommonVars(request_rec *r)
     {
     table *e=r->subprocess_env;
     int keysize=0;
     int secretkeysize=0;
     char *t;
-    char *cipher;
+    const char *cipher;
     int n;
     X509 *xs;
 #if APACHE_SSL_EXPORT_CERTS && !APACHE_SSL_KEEP_CERTS
@@ -677,8 +741,9 @@ static int SSLFixups(request_rec *r)
 
     /* Check to see if SSL is on */
     if(pConfig->bDisabled || OPTIONAL_SSL)
-	return DECLINED;
+	return;
 
+    if (r->connection->client->ssl == NULL) return;
     cipher=SSL_get_cipher(r->connection->client->ssl);
     ap_table_set(e,"HTTPS","on");
     ap_table_set(e,"HTTPS_CIPHER",cipher);
@@ -703,16 +768,29 @@ static int SSLFixups(request_rec *r)
 
     if(r->connection->client->szClientX509)
 	{
-	ExpandCert(r->pool,e,"SSL_CLIENT_",
+	ExpandCert(r->pool,e,"SSL_CLIENT_","SSL_CLIENT_DN",
 		   r->connection->client->szClientX509);
+#ifdef USE_EXTENDED_ENVIRONMENT
+	ExpandCert(r->pool,e,"SSL_CLIENT_S_DN_","SSL_CLIENT_S_DN",
+		   r->connection->client->szClientX509);
+#endif
 	xs=SSL_get_peer_certificate(r->connection->client->ssl);
 	ExpandCert(r->pool,e,
 #if USE_OLD_ENVIRONMENT
-		   "SSL_CLIENT_I",
+		   "SSL_CLIENT_I","SSL_CLIENT_I_DN",
 #else
-		   "SSL_CLIENT_I_",
+		   "SSL_CLIENT_I_","SSL_CLIENT_I_DN",
 #endif
 		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+#ifdef USE_EXTENDED_ENVIRONMENT
+	ExpandCert(r->pool,e,"SSL_CLIENT_I_DN_","SSL_CLIENT_I_DN",
+		   X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+	get_serial(r->pool,e,"SSL_CLIENT_M_SERIAL",xs);
+        get_valid(r->pool,e,"SSL_CLIENT_V_START",X509_get_notBefore(xs));
+        get_valid(r->pool,e,"SSL_CLIENT_V_END",X509_get_notAfter(xs));
+        t=ap_psprintf(r->pool,"%lu",X509_get_version(xs)+1);
+        ap_table_setn(e,"SSL_CLIENT_M_VERSION",t);
+#endif
 #if APACHE_SSL_EXPORT_CERTS
 	if(pDirConfig->bExportCerts)
 	    AddCertToEnv(r->pool,e,"SSL_CLIENT_CERT",xs);
@@ -720,15 +798,29 @@ static int SSLFixups(request_rec *r)
 	}
 
     xs=SSL_get_certificate(r->connection->client->ssl);
-    ExpandCert(r->pool,e,"SSL_SERVER_",
+    ExpandCert(r->pool,e,"SSL_SERVER_","SSL_SERVER_DN",
 	       X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
+#ifdef USE_EXTENDED_ENVIRONMENT
+    ExpandCert(r->pool,e,"SSL_SERVER_S_","SSL_SERVER_S_DN",
+               X509_NAME_oneline(X509_get_subject_name(xs),NULL,0));
+#endif
     ExpandCert(r->pool,e,
 #if USE_OLD_ENVIRONMENT
-	       "SSL_SERVER_I",
+	       "SSL_SERVER_I","SSL_SERVER_I_DN",
 #else
-	       "SSL_SERVER_I_",
+	       "SSL_SERVER_I_","SSL_SERVER_I_DN",
 #endif
 	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+#ifdef USE_EXTENDED_ENVIRONMENT
+    ExpandCert(r->pool,e,
+	       "SSL_SERVER_I_DN_","SSL_SERVER_I_DN",
+	       X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0));
+    get_serial(r->pool,e,"SSL_SERVER_M_SERIAL",xs);
+    get_valid(r->pool,e,"SSL_SERVER_V_START",X509_get_notBefore(xs));
+    get_valid(r->pool,e,"SSL_SERVER_V_END",X509_get_notAfter(xs));
+    t=ap_psprintf(r->pool,"%lu",X509_get_version(xs)+1);
+    ap_table_setn(e,"SSL_SERVER_M_VERSION",t);
+#endif
 
 #if APACHE_SSL_EXPORT_CERTS
 # if !APACHE_SSL_KEEP_CERTS
@@ -746,8 +838,16 @@ static int SSLFixups(request_rec *r)
 # endif
 #endif
 
+    return;
+    }
+
+/*
+static int SSLFixups(request_rec *r)
+    {
+    SSLAddCommonVars(r);
     return DECLINED;
     }
+*/
 
 static void *CreateSSLServerConfig(pool *p,server_rec *s)
     {
@@ -774,7 +874,7 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     return rec;
     }
 
-#define Merge(el,unset)		new->el=add->el == unset ? base->el : add->el
+#define Merge(el,unset)		new->el=(add->el == unset ? base->el : add->el)
 #define MergeString(el)		Merge(el,NULL)
 #define MergeBool(el)		Merge(el,UNSET)
 #define MergeInt(el)		Merge(el,0)
@@ -808,7 +908,7 @@ static void *CreateSSLDirConfig(pool *p,char *dummy)
 
     rec->tbRequiredCiphers=ap_make_table(p,4);
     rec->tbBannedCiphers=ap_make_table(p,4);
-    rec->bSSLRequired=FALSE;
+    rec->nSSLRequired=SSLREQ_UNSET;
 #if APACHE_SSL_EXPORT_CERTS
     rec->bExportCerts=FALSE;
 #endif
@@ -831,7 +931,7 @@ static void *MergeSSLDirConfig(pool *p,void *basev,void *addv)
 					     base->tbRequiredCiphers);
     new->tbBannedCiphers=ap_overlay_tables(p,add->tbBannedCiphers,
 					   base->tbBannedCiphers);
-    new->bSSLRequired=base->bSSLRequired || add->bSSLRequired;
+    Merge(nSSLRequired,SSLREQ_UNSET);
 #if APACHE_SSL_EXPORT_CERTS
     new->bExportCerts=base->bExportCerts || add->bExportCerts;
 #endif
@@ -935,6 +1035,9 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
 	t=GlobalCacheGet(aucKey,nKey,pnData,&tExpiresAt);
 	if(!t)
 	    return NULL;
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session found in global cache, to=%ld\n", tExpiresAt);
+#endif
 
 	t2=malloc(nKey);
 	memcpy(t2,aucKey,nKey);
@@ -947,6 +1050,9 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
 	t=p->aucData;
 	*pnData=p->nData;
 	*pbLocal=TRUE;
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session found in local cache\n");
+#endif
 	}
     return t;
     }
@@ -1035,6 +1141,7 @@ static void SendSessionToServer(SSL_SESSION *pSession,
     pConfig=ap_get_module_config(pCurrentConnection->server->module_config,
 				 &apache_ssl_module);
     tExpiresAt=time(NULL)+pConfig->tSessionCacheTimeout;
+    SSL_set_timeout(pSession, pConfig->tSessionCacheTimeout);
 
     t=malloc(pSession->session_id_length);
     memcpy(t,pSession->session_id,pSession->session_id_length);
@@ -1069,7 +1176,12 @@ static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength,
 
     t=CacheFind(aucSessionID,nLength,&nData,&bLocal);
     if(!t)
+	{
+#if DEBUG_SESSIONS
+	fprintf(stderr,"Session not found\n");
+#endif
 	return NULL;
+	}
 
     pSession=d2i_SSL_SESSION(NULL,&t,nData);
     if(!pSession)
@@ -1149,7 +1261,8 @@ static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
 	}
 
     pConfig->prsaKey=RSA_new();
-    for(nCount=0 ; !PEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL) ; ++nCount)
+    for(nCount=0 ; !localPEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL)
+	  ; ++nCount)
 	{
 	if(nCount < 3 && ERR_peek_error() == ERR_PACK(ERR_LIB_EVP,
 						      EVP_F_EVP_DECRYPTFINAL,
@@ -1170,7 +1283,7 @@ static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
 		  (const char *)pConfig->prsaKey);
     }
 
-#if SSLEAY_VERSION_NUMBER < 0x0920 || RENEG
+#if SSLEAY_VERSION_NUMBER < 0x0920
 static int X509NameCompare(X509_NAME **a,X509_NAME **b)
     {
     return X509_NAME_cmp(*a,*b);
@@ -1182,7 +1295,7 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
     FILE *f;
     char szPath[MAX_STRING_LEN];
 #if SSLEAY_VERSION_NUMBER >= 0x0920
-    STACK *stackCAs=NULL;
+    STACK_OF(X509_NAME) *stackCAs=NULL;
 #endif
 
     if(pConfig->px509Certificate)
@@ -1255,7 +1368,7 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 
     pConfig->px509Certificate=X509_new();
 
-    if(!PEM_read_X509(f,&pConfig->px509Certificate,NULL))
+    if(!localPEM_read_X509(f,&pConfig->px509Certificate,NULL))
 	{
 	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
 		     "Error reading server certificate file %s",szPath);
@@ -1286,7 +1399,7 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 #endif
 	{
 	if(pConfig->szCACertificateFile || pConfig->szCACertificatePath)
-	    stackCAs=sk_new(NULL);
+	    stackCAs=sk_X509_NAME_new(NULL);
 
 	if(pConfig->szCACertificateFile
 	   && !SSL_add_file_cert_subjects_to_stack(stackCAs,
@@ -1304,7 +1417,7 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 	    exit(1);
 	    }
 
-	if(stackCAs && sk_num(stackCAs) > 0)
+	if(stackCAs && sk_X509_NAME_num(stackCAs) > 0)
 	    SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,stackCAs);
 	}
 #elif SSLEAY_VERSION_NUMBER >= 0x0800
@@ -1592,7 +1705,7 @@ static void CloseSSL(void *_conn)
     }
 #endif
 
-int ApacheSSLSetupConnection(conn_rec *conn)
+static int SSLSetupConnection(conn_rec *conn)
     {
     server_rec *srvr=conn->server;
     BUFF *fb=conn->client;
@@ -1696,6 +1809,9 @@ int ApacheSSLSetupConnection(conn_rec *conn)
     return TRUE;
     }
 
+static void SSLSetupConnection2(conn_rec *conn)
+    { SSLSetupConnection(conn); }
+
 static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
     {
     return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
@@ -1729,7 +1845,14 @@ static const char *SSLBanCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
 static const char *SSLRequireSSL(cmd_parms *cmd,SSLDirConfigRec *rec,
 				 char *cipher)
     {
-    rec->bSSLRequired=TRUE;
+    rec->nSSLRequired=SSLREQ_REQUIRED;
+    return NULL;
+    }
+
+static const char *SSLDenySSL(cmd_parms *cmd,SSLDirConfigRec *rec,
+				 char *cipher)
+    {
+    rec->nSSLRequired=SSLREQ_DENIED;
     return NULL;
     }
 
@@ -1788,7 +1911,7 @@ static const char *SSLRandomFilePerConnection(cmd_parms *cmd,char *struct_ptr,
 					      char *arg1,char *arg2)
     {
     SSLConfigRec *pConfig=ap_get_module_config(cmd->server->module_config,
-						  &apache_ssl_module);
+					       &apache_ssl_module);
     RandomFile *prf=ap_push_array(pConfig->aRandomFiles);
 
     prf->szFile=arg1;
@@ -1898,6 +2021,8 @@ command_rec ssl_cmds[]=
       "Add a cipher to the per directory list of banned ciphers" },
     { "SSLRequireSSL",SSLRequireSSL,NULL,OR_FILEINFO,NO_ARGS,
       "Require SSL (use to protect directories that should not be served unencrypted)" },
+    { "SSLDenySSL",SSLDenySSL,NULL,OR_FILEINFO,NO_ARGS,
+      "Deny SSL (use to constrain directories that should be served unencrypted)" },
 #if APACHE_SSL_EXPORT_CERTS
     { "SSLExportClientCertificates",SSLExportCerts,NULL,OR_FILEINFO,NO_ARGS,
       "Export client certificates and their certificate chain" },
@@ -1920,12 +2045,12 @@ command_rec ssl_cmds[]=
       "A file to read randomness from and a maximim number of bytes to read on each request" },
     { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
       "Defunct SSL log file - DO NOT USE" },
-    { NULL },
+    { NULL }
     };
 
 module MODULE_VAR_EXPORT apache_ssl_module =
     {
-    STANDARD_MODULE_STUFF,
+    SSL_MODULE_STUFF,
     InitSSLServer,		/* initializer */
     CreateSSLDirConfig,		/* dir config creater */
     MergeSSLDirConfig,		/* dir merger --- default is to override */
@@ -1938,7 +2063,13 @@ module MODULE_VAR_EXPORT apache_ssl_module =
     NULL,			/* check auth */
     SSLCheckCipher,		/* check access */
     NULL,			/* type_checker */
-    SSLFixups,			/* fixups */
+    /*    SSLFixups,			fixups */
+    NULL,			/* fixups */
     NULL,			/* logger */
-    }
-;
+    NULL,			/* header_parser */
+    NULL,			/* child_init */
+    NULL,			/* child_exit */
+    NULL,			/* post_read_request */
+    SSLSetupConnection2,	/* setup_connection */
+    SSLAddCommonVars		/* add_common_vars */
+    };
